//*****************************************************************************
//   Copyright (c) 1998 INRIA.
//   http://www.inria.fr/croap/aioli/modules/figue/web/credits.html#COPYRIGHT
//*****************************************************************************

package figue;

import figue.box.GlyphInterface;
import figue.box.RootGlyph;
import figue.path.PathInterface;
import figue.path.SelectionPathInterface;
import figue.path.SelectionPath;
import figue.resource.ColorInterface;
import figue.resource.WrongResourceException;
import figue.resource.BasicColor;
import figue.resource.FontInterface;
import figue.resource.FontFamily;
import figue.resource.FontStyle;
import figue.resource.PortableFontInterface;
import figue.resource.StyleInterface;
import figue.resource.ResourceFactoryInterface;
import figue.selection.SelectionInterface;
import figue.selection.WrongSelectionException;
import figue.selection.Selection;
import figue.geometry.Dimension;
import figue.geometry.Rectangle;
import figue.format.AlternativeInterface;
import figue.format.Alternative;
import figue.format.ImmediateDistance;
import figue.error.ErrorHandler;
import figue.error.ErrorHandlerInterface;
import figue.error.Warning;
import figue.error.Error;
import figue.error.ImplementationError;
import figue.multithread.ThreadConditionInterface;
import figue.multithread.ThreadCondition;
import figue.multithread.NotifyingThreadStateInterface;
import figue.multithread.StoppedFacadeException;
import figue.multithread.NotifyingThreadState;
import figue.io.xml.FigueLexer;
import figue.io.xml.FigueParser;
import figue.io.xml.FigueWalker;
import figue.io.Utility;
import figue.internal.AvailableVersion;
import figue.internal.BooleanWrapper;
import figue.internal.Cursor;
import figue.internal.Globals;
import figue.internal.GlyphViewerInterface;
import figue.internal.PathBuildingVisitor;
import figue.internal.PathModificationVisitor;
import figue.internal.PathSelectionVisitor;
import figue.internal.PathVisitorInterface;
import figue.internal.StackInterface;

import java.util.Hashtable;
import java.io.IOException;
import java.io.Writer;
import java.util.Enumeration;
import java.util.Vector;
import java.io.InputStream;
import java.io.IOException;
import antlr.ParserException;
import antlr.CommonAST;

/**
 * [PENDING: update the comments] Facade.
 */

public abstract class Facade extends ErrorHandler implements FacadeInterface {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/Facade.java,v $ $revision$ $Date: 1998/11/02 15:54:28 $ Copyright 1998 INRIA.";
    
    interface ThreadInterface extends Runnable {
	public void awake();
    }
    
    abstract class Thread extends java.lang.Thread implements ThreadInterface {
	public Thread(String aName,
		      ThreadConditionInterface aCondition,
		      int aPriority) {
	    super(aName);
	    _condition = aCondition;
	    setPriority(aPriority);
	}
	protected abstract boolean isInRightState() throws StoppedFacadeException;
	protected abstract void execute();
	public final void run() {
	    try {
		synchronized ( _condition ) {
		    while ( ! isDone() ) {
			if ( isInRightState() ) {
			    execute();
			}
			else {
			    _condition.await();
			}
		    }
		}
	    }
	    catch ( StoppedFacadeException anException ) {
		;
	    }
	}
	public final void awake() {
	    _condition.signal();
	}
	public final ThreadConditionInterface getCondition() {
	    return _condition;
	}
	private ThreadConditionInterface _condition;
    }
  
    final class FormatThread extends Thread {
	FormatThread(ThreadConditionInterface aCondition) {
	    super("Format",aCondition,Globals.FORMAT_THREAD_PRIORITY);
	}
	public final boolean isInRightState() throws StoppedFacadeException {
	    return isFormatting();
	}
	public final void execute() {
	    executeFormatTask();
	}
	public final String toString() {
	    return "FormatThread<>";
	}
    }

    final class BuildThread extends Thread {
	public BuildThread(ThreadConditionInterface aCondition) {
	    super("Build",aCondition,Globals.BUILD_THREAD_PRIORITY);
	}
	protected final boolean isInRightState() throws StoppedFacadeException {
	    return isBuilding();
	}
	public final void execute() {
	    executeBuildTask(getCondition()); //// TEMPO
	}
	public final String toString() {
	    return "BuildThread<>";
	}
    }
  
    protected Facade(AvailableVersion aVersion,
		     FacadeInitializer anInitializer) {
	_disposedFlag = false;
	if ( anInitializer.getFormattingWidth() > 0 ) {
	    _formattingWidth = anInitializer.getFormattingWidth();
	}
	else {
	    _formattingWidth = aVersion.getDefaultFormattingWidth();
	}
	_version = aVersion;
	_formatable = new ThreadCondition("Formatable");
	_formatSuspended = new ThreadCondition("Format suspended");
	_drawable = new ThreadCondition("Drawable");
	_building = new ThreadCondition("Building");
	_inactive = new ThreadCondition("Inactive");
	_threadState = new NotifyingThreadState(this);
	_formatThread = new FormatThread(new ThreadCondition("Formatting"));
	_buildThread = new BuildThread(new ThreadCondition("Building glyphs"));
	_rootGlyph = new RootGlyph(this);
	_selections = new Hashtable(Globals.AVERAGE_SELECTION_NUMBER,
				    Globals.VERY_FAST_ACCESS);

	_rootSelectionPath = new RootSelectionPath();
	_buildPathQueue = new Vector();
	_buildFrozen = false;
	_rectangleCacheForRefresh = new Rectangle();
	_rectangleCacheForRefreshVisible = new Rectangle();
	_emptyRectangle = new Rectangle();
	_booleanCacheForRefresh = new BooleanWrapper();
	_rectangleCacheForRefreshWork = new Rectangle();
	_alternative = new Alternative();
	_pathModificationVisitor = new PathModificationVisitor(this);
	_pathBuildingVisitor = new PathBuildingVisitor(this);
	_pathSelectionVisitor = new PathSelectionVisitor(this);
	_distance = new ImmediateDistance(-1);
	_factory = aVersion.newResourceFactory(this,anInitializer.getPortableFontLibrary());
	_errorHandler = _factory.newErrorHandler();
	_formatThread.start();
	_buildThread.start();
    }

    protected Facade(AvailableVersion aVersion,
		     FacadeInitializer anInitializer,
		     InputStream anInput) throws ParserException, IOException {
	this(aVersion,anInitializer);
	final FigueLexer theLexer = new FigueLexer(anInput);
	final FigueParser theParser = new FigueParser(theLexer);
	theParser.document();
	final CommonAST theTree = (CommonAST)theParser.getAST();
	//	System.out.println("=> " + theTree.toStringList() );
    	final FigueWalker theWalker = new FigueWalker();
    	theWalker.facade(theTree,this);
    }

    protected Facade(AvailableVersion aVersion,
		     FacadeInitializer anInitializer,
		     FacadeInterface aFacade) throws WrongResourceException, CloneNotSupportedException {
	this(aVersion,anInitializer);
	ColorInterface theColor;
	for ( Enumeration e = aFacade.colors(); e.hasMoreElements(); ) {
	    theColor = (ColorInterface)(e.nextElement());
	    if ( ! isColorDefined(theColor.getName()) ) {
		doGetColor(theColor.getName(),theColor.getRed(),theColor.getGreen(),theColor.getBlue());
	    }
	}
	FontInterface theFont;
	for ( Enumeration e = aFacade.fonts(); e.hasMoreElements(); ) {
	    theFont = (FontInterface)(e.nextElement());
	    if ( ! isFontDefined(theFont.getName()) ) {
		doGetFont(theFont.getName(),theFont.getFamily(),theFont.getStyle(),theFont.getSize(),theFont.isUnderlined());
	    }
	}
	StyleInterface theStyle;
	ColorInterface theBackground;
	for ( Enumeration e = aFacade.styles(); e.hasMoreElements(); ) {
	    theStyle = (StyleInterface)(e.nextElement());
	    if ( ! isStyleDefined(theStyle.getName()) ) {
		if ( theStyle.getFont() != null ) {
		    theFont = getFont(theStyle.getFont().getName());
		}
		else {
		    theFont = null;
		}
		if ( theStyle.getColor() != null ) {
		    theColor = getColor(theStyle.getColor().getName());
		}
		else {
		    theColor = null;
		}
		if ( theStyle.getBackground() != null ) {
		    theBackground = getColor(theStyle.getBackground().getName());
		}
		else {
		    theBackground = null;
		}
		doGetStyle(theStyle.getName(),theFont,theColor,theBackground);
	    }
	}
	if ( aFacade.getRootGlyph().numberOfChildren() > 0 ) {
	    _rootGlyph.addChild((GlyphInterface)(aFacade.getRootGlyph().getChild(0).clone()));
	}
    }
    
    protected void init() {
	_pathModificationVisitor.init();
	_pathBuildingVisitor.init();
	_pathSelectionVisitor.init();
    }
    
    public final int getFormattingWidth() {
	return _formattingWidth;
    }

    public final void setFormattingWidth(int aWidth) {
	try {
	    synchronized ( _inactive ) {
		while ( true ) {
		    synchronized ( this ) {
			if ( _threadState.isInactive() ) {
			    if ( ( getRootGlyph().numberOfChildren() == 0 ) || 
				 ( _formattingWidth == aWidth ) ) {
				return;
			    }
			    else {
				_formattingWidth = aWidth;
			    }
			    getRootGlyph().setNew();
			    getRootGlyph().getChild(0).setNew();
			    getContext().setCursor(Cursor.WAIT);
			    _distance.setValue(getFormattingWidth());
			    getVersion().reFormat(_rootGlyph,
						  getContext().getFontMetric(),
						  _alternative,
						  _distance);
			    break;
			}
		    }
		    _inactive.await();
		}
	    }
	    getContext().touch();
	}
	catch ( StoppedFacadeException anException ) {
	    return;
	}
    }

    public final void updateAndRedisplay(PathInterface aPath) {
	getViewer().updateAndRedisplay(aPath);
    }

    private final void firstGlyphStructureUpdate(PathInterface aPath) {
	if ( aPath != null ) {
	    aPath.accept(_pathBuildingVisitor,_rootGlyph,_rootSelectionPath);
	    StackInterface theEmptyLeaves = _pathBuildingVisitor.getEmptyLeaves();
	    while ( ! theEmptyLeaves.isEmpty() ) {
		((SelectionPathInterface)theEmptyLeaves.pop()).cutEmptyLeaf();
	    }
	    getContext().buildAdvance();
	}
    }

    final Rectangle interactiveUpdateGlyphStructure(PathInterface aPath,
						    Rectangle aVisibleZone) {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> interactiveUpdateGlyphStructure");
	}
	Rectangle theRefreshZone;
	try {
	    synchronized ( _inactive ) {
		while ( true ) {
		    synchronized ( this ) {
			if ( _threadState.isInactive() ) {
			    theRefreshZone = firstGlyphStructureUpdate(aPath,aVisibleZone);
			    if ( ( _rootGlyph != null ) &&
				 ( _rootGlyph.numberOfChildren() == 1 ) &&
				 ( ! _rootGlyph.isClean() ) ) {
				if ( Globals.DEBUG_MULTI_THREAD ) {
				    System.out.println("\nFORMAT TEMPO\n");
				}
				_distance.setValue(getFormattingWidth());
				getVersion().reFormat(_rootGlyph,
						      getContext().getFontMetric(),
						      _alternative,
						      _distance);
			    }
			    break;
			}
		    }
		    _inactive.await();
		}
	    }
	}
	catch ( StoppedFacadeException anException ) {
	    return null;
	}
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("<- interactiveUpdateGlyphStructure ");
	}
	return computeRefreshZone(aVisibleZone,theRefreshZone);
    }

    private final Rectangle computeRefreshZone(Rectangle aVisibleZone,
					       Rectangle aRefreshZone) {
	_booleanCacheForRefresh.setTrue();
	_rectangleCacheForRefreshVisible.setBounds(aVisibleZone);
	_rootGlyph.cleanModificationAndUpdateRefreshZone(_rectangleCacheForRefreshVisible,
							 _rectangleCacheForRefresh,
							 _booleanCacheForRefresh,
							 _rectangleCacheForRefreshWork,
							 getContext());
	if ( aRefreshZone.isEmpty() ) {
	    aRefreshZone.setBounds(_rectangleCacheForRefresh);
	}
	else {
	    aRefreshZone.setBounds(aRefreshZone.union(_rectangleCacheForRefresh));
	}
	return aRefreshZone;
    }

    private final Rectangle firstGlyphStructureUpdate(PathInterface aPath,
						      Rectangle aVisibleZone) {
	_pathModificationVisitor.setVisibleZone(aVisibleZone);
	aPath.accept(_pathModificationVisitor,_rootGlyph,_rootSelectionPath);
	StackInterface theEmptyLeaves = _pathModificationVisitor.getEmptyLeaves();
	while ( ! theEmptyLeaves.isEmpty() ) {
	    ((SelectionPathInterface)theEmptyLeaves.pop()).cutEmptyLeaf();
	}
	return _pathModificationVisitor.getRefreshZone();
    }
    
    public final void buildInit(PathInterface aPath) {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> buildInit ");
// 	    if ( aPath != null ) {
// 		try {
// 		    java.io.PrintWriter theOutput = new java.io.PrintWriter(System.out);
// 		    aPath.writeState(theOutput);
// 		    theOutput.flush();
// 		}
// 		catch ( java.io.IOException e ) {
// 		    e.printStackTrace();
// 		    System.exit(0);
// 		}
// 	    }
	}
	try {
	    synchronized ( _inactive ) {
		while ( true ) {
		    synchronized ( this ) {
			if ( _threadState.isInactive() ) {
			    getContext().setCursor(Cursor.WAIT);
			    _buildFrozen = false;
			    _threadState.startUpdateGlyph();
			    if ( aPath != null ) { /// +++
				firstGlyphStructureUpdate(aPath);
			    }
			    _threadState.endUpdateGlyph();
			    break;
			}
		    }
		    _inactive.await();
		}
	    }
	    awakeFormatting();
	    getContext().touch();
	}
	catch ( StoppedFacadeException anException ) {
	    System.exit(0);
	}
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("<- buildInit");
	}
    }

    public final void build(PathInterface aPath) {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> build ");
// 	    if ( aPath != null ) {
// 		try {
// 		    java.io.PrintWriter theOutput = new java.io.PrintWriter(System.out);
// 		    aPath.writeState(theOutput);
// 		    theOutput.flush();
// 		}
// 		catch ( java.io.IOException e ) {
// 		    e.printStackTrace();
// 		    System.exit(0);
// 		}
// 	    }
	}
	try {
	    synchronized ( _building ) {
		while ( true ) {
		    synchronized ( this ) {
			if ( Globals.ROBUST ) {
			    if ( _buildFrozen ) {
				throw new ImplementationError("Unexpected");
			    }
			}
			if ( _threadState.isBuilding() ) {
			    _buildPathQueue.insertElementAt(aPath,_buildPathQueue.size());
			    break;
			}
		    }
		    _building.await();
		}
	    }
	    awakeBuilding();
	}
	catch ( StoppedFacadeException anException ) {
	    return;
	}
    }

    public final void buildComplete() {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> buildComplete");
	}
	try {
	    synchronized ( _building ) {
		if ( Globals.ROBUST ) {
		    if ( ! _threadState.isBuilding() ) {
			throw new ImplementationError("Unexpected");
		    }
		    if ( _buildFrozen ) {
			throw new ImplementationError("Unexpected");
		    }
		}
		synchronized ( this ) {
		    getContext().setCursor(Cursor.TEXT);
		    _buildFrozen = true;
		}
		getContext().buildComplete();
	    }
	    awakeBuilding();
	}
	catch ( StoppedFacadeException anException ) {
	    return;
	}
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("<- buildComplete");
	}
    }
    
    public final void updateSelection(PathInterface aPath) {
	try {
	    _emptyRectangle.setBounds(0,0,0,0);
	    interactiveUpdateSelection(aPath,_emptyRectangle);
	}
	catch ( StoppedFacadeException anException ) {
	    return;
	}
    }

    public final void updateSelectionAndRedisplay(PathInterface aPath) {
	getViewer().updateSelectionAndRedisplay(aPath);
    }

    final Rectangle interactiveUpdateSelection(PathInterface aPath,
					       Rectangle aVisibleZone)
	throws StoppedFacadeException {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> updateSelection");
	}
	synchronized ( _inactive ) {
	    while ( true ) {
		synchronized ( this ) {
		    if ( _threadState.isInactive() ) {
			_pathSelectionVisitor.setVisibleZone(aVisibleZone);
			
			aPath.accept(_pathSelectionVisitor,_rootGlyph,_rootSelectionPath);
			StackInterface theEmptyLeaves = _pathSelectionVisitor.getEmptyLeaves();
			while ( ! theEmptyLeaves.isEmpty() ) {
			    ((SelectionPathInterface)theEmptyLeaves.pop()).cutEmptyLeaf();
			}
			break;
		    }
		}
		_inactive.await();
	    }
	}
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("<- updateSelection");
	}
	return _pathSelectionVisitor.getRefreshZone();
    }

    abstract GlyphViewerInterface getViewer();

    public final void executeWhenInactive(CommandInterface aCommand) {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> executeWhenInactive");
	    System.out.flush();
	}
	try {
	    synchronized ( _inactive ) {
		while ( true ) {
		    synchronized ( this ) {
			if ( _threadState.isInactive() ) {
			    if ( aCommand.execute() ) {
				break;
			    }
			}
		    }
		    _inactive.await();
		}
	    }
	    awake();
	}
	catch ( StoppedFacadeException anException ) {
	    return;
	}
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("<- executeWhenInactive");
	    System.out.flush();
	}
    }

    public void suspendFormatting() {
	synchronized ( this ) {
	    _threadState.suspendFormat();
	}
	awakeDrawable();
    }

    public final void waitUntilFormattingResumed() {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> waitUntilFormattingResumed ");
	}
	try {
	    synchronized ( _formatSuspended ) {
		while ( true ) {
		    synchronized ( this ) {
			if ( _threadState.isFormatting() ) {
			    break; 
			}
		    }
		    _formatSuspended.await();
		}
	    }
	}
	catch ( StoppedFacadeException anException ) {
	    return;
	}
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("<- waitUntilFormattingResumed ");
	}
    }
    
    public final void insureVisible(PathInterface aPath) {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> insureVisible");
	}
	try {
	    synchronized ( _inactive ) {
		while ( true ) {
		    synchronized ( this ) {
			if ( _threadState.isInactive() ) {
			    insureVisibleWhenInactive(aPath);
			    if ( Globals.DEBUG_MULTI_THREAD ) {
				System.out.println("<- insureVisible");
			    }
			    return;
			}
		    }
		    _inactive.await();
		}
	    }
	}
	catch ( StoppedFacadeException anException ) {
	    return;
	}
    }

    protected abstract void insureVisibleWhenInactive(PathInterface aPath);

    private final void markAllDirty(GlyphInterface aGlyph) {
	aGlyph.setDirty();
	final int theSize = aGlyph.numberOfChildren();
	for( int i = 0; i < theSize; i++ ) {
	    markAllDirty(aGlyph.getChild(i));
	}
    }

    final Dimension getRootGlyphSizeEstimation() {
	return _rootGlyph.getSizeEstimation();
    }

    public final boolean isSelectionDefined(String aMnemonic) {
	return _selections.containsKey(aMnemonic);
    }
  
    public final SelectionInterface getSelection(String aName) 
	throws WrongSelectionException {
	SelectionInterface theSelection = (SelectionInterface)(_selections.get(aName));
	if ( theSelection != null ) {
	    return theSelection;
	}
	throw new WrongSelectionException("Unknown selection: " + aName);
    }
  
    public final SelectionInterface doGetSelection(String aName,
						   Priority aPriority,
						   String aColorName,
						   String aBackgroundName)
	throws WrongSelectionException {
	if ( _selections.containsKey(aName) ) {
	    throw new WrongSelectionException("Selection name already in use: " + 
					      aName);
	}
	SelectionInterface theSelection = new Selection(aName,aPriority,aColorName,aBackgroundName,this);
	_selections.put(aName,theSelection);
	return theSelection;      
    }

    public final SelectionInterface doGetSelection(String aName,
						   Priority aPriority,
						   ColorInterface aColor,
						   ColorInterface aBackground) 
	throws WrongSelectionException {
	if ( _selections.containsKey(aName) ) {
	    throw new WrongSelectionException("Selection name already in use: " + 
					      aName);
	}
	SelectionInterface theSelection = new Selection(aName,aPriority,aColor,aBackground,this);
	_selections.put(aName,theSelection);
	return theSelection;
    }

    public final Enumeration selections() {
	return _selections.elements();
    }

    public final ColorInterface doGetDefaultColor() {
	return _factory.doGetDefaultColor();
    }
    
    public final ColorInterface doGetDefaultBackground() {
	return _factory.doGetDefaultBackground();
    }

    public final boolean isColorDefined(String aMnemonic) {
	return _factory.isColorDefined(aMnemonic);
    }

    public final ColorInterface getColor(String aMnemonic) 
	throws WrongResourceException {
	return _factory.getColor(aMnemonic);
    }
  
    public final ColorInterface doGetColor(String aMnemonic,BasicColor aColor)
	throws WrongResourceException {
	return _factory.doGetColor(aMnemonic,aColor);
    }

    public final ColorInterface doGetColor(String aMnemonic,
					   int red, int green, int blue)  
	throws WrongResourceException {
	return _factory.doGetColor(aMnemonic,red,green,blue);
    }

    public final Enumeration colors() {
	return _factory.colors();
    }
  
    public final FontInterface doGetDefaultFont() {
	return _factory.doGetDefaultFont();
    }

    public final boolean isFontDefined(String aMnemonic) {
	return _factory.isFontDefined(aMnemonic);
    }

   public final FontInterface getFont(String aMnemonic) 
       throws WrongResourceException {
       return _factory.getFont(aMnemonic);
   }

   public final FontInterface doGetFont(String aMnemonic,
					FontFamily aFamily,
					FontStyle aStyle,
					int aPointSize,
					boolean anUnderlinedFlag) 
       throws WrongResourceException {
       return _factory.doGetFont(aMnemonic,aFamily,aStyle,aPointSize,anUnderlinedFlag);
   }

   public final Enumeration fonts() {
       return _factory.fonts();
   }

   public final StyleInterface doGetDefaultStyle() {
       return _factory.doGetDefaultStyle();
   }

   public final boolean isStyleDefined(String aMnemonic) {
       return _factory.isStyleDefined(aMnemonic);
   }

   public final StyleInterface getStyle(String aMnemonic) 
       throws WrongResourceException {
       return _factory.getStyle(aMnemonic);
   }

   public final StyleInterface doGetStyle(String aMnemonic,
					  FontInterface aFont,
					  ColorInterface aColor,
					  ColorInterface aBackground) 
       throws WrongResourceException {
       return _factory.doGetStyle(aMnemonic,aFont,aColor,aBackground);
   }

   public final Enumeration styles() {
       return _factory.styles();
    }

   public final boolean isPortableFontDefined(String aMnemonic) {
       return _factory.isPortableFontDefined(aMnemonic);
   }

   public final PortableFontInterface getPortableFont(String aMnemonic) 
       throws WrongResourceException {
       return _factory.getPortableFont(aMnemonic);
   }

    public final PortableFontInterface doGetPortableFont(String aMnemonic,
							 String aName,
							 int aDpi,
							 double aScale) 
       throws WrongResourceException {
       return _factory.doGetPortableFont(aMnemonic,aName,aDpi,aScale);
   }

   public final Enumeration portableFonts() {
       return _factory.portableFonts();
   }

   public final SelectionPathInterface getSelectionPath() {
       return _rootSelectionPath;
   }

   protected final AvailableVersion getVersion() {
       return _version;
   }
  
    public final GlyphInterface getRootGlyph() { /// TEMPO
       return _rootGlyph;
   }

   private final Vector getBuildPathQueue() {
       return _buildPathQueue;
   }

   protected final NotifyingThreadStateInterface getThreadState() {
       return _threadState;
   }

   private final boolean isDone() {
       return getThreadState().isDone();
   }

   private final boolean isDrawing() throws StoppedFacadeException {
       return getThreadState().isDrawing();
   }
  
   private final boolean isFormatting() throws StoppedFacadeException {
       return getThreadState().isFormatting();
   }
  
   private final boolean isBuilding() throws StoppedFacadeException {
       return getThreadState().isBuilding();
   }

   protected final void endDraw() {
       getThreadState().endDraw();
   }

   private final void endFormat() {
       getThreadState().endFormat();
   }
  
   private final void endBuild() {
       getThreadState().endBuild();
   }

   protected final ThreadConditionInterface getDrawable() {
       return _drawable;
   }

   protected final ThreadConditionInterface getFormatable() {
       return _formatable;
   }

   protected final ThreadConditionInterface getInactive() {
       return _inactive;
   }

   private final void awakeInactive() {
       _inactive.signal();
   }

   private final void awakeFormatable() {
       _formatable.signal();
   }

   protected final void awakeFormatSuspended() {
       _formatSuspended.signal();
   }

   private final void awakeDrawable() {
       _drawable.signal();
   }

   protected final void awakeFormatting() {
       _formatThread.awake();
   }

   private final void awakeBuilding() {
       _buildThread.awake();
   }

    public final void awake() { /// +++ NEW PUBLIC
       awakeFormatSuspended();
       awakeDrawable();
       awakeFormatable();
       _building.signal();
       awakeInactive();
   }

   public final void awakeWithoutInactive() { //// +++ NEW PUBLIC BEURK
       awakeFormatSuspended();
       awakeDrawable();
       awakeFormatable();
       _building.signal();
   }

   public final boolean hasNextErrorHandler() {
       return ( _errorHandler != null );
   }

   public final ErrorHandlerInterface getNextErrorHandler() {
       return _errorHandler;
   }

   public final boolean canHandleError() {
       return false;
   }

   protected abstract void executeDrawTask();

   protected final void executeFormatTask() {
       GlyphInterface theRoot = getRootGlyph();
       if ( ( theRoot != null ) &&
	    ( theRoot.numberOfChildren() == 1 ) &&
	    ( ! theRoot.isClean() ) ) {
	   if ( Globals.DEBUG_MULTI_THREAD ) {
	       System.out.println("\nbegin executeFormatTask format\n");
	   }
	   _distance.setValue(getFormattingWidth());
	   getVersion().format(theRoot,getContext().getFontMetric(),_alternative,_distance);
	   if ( Globals.DEBUG_MULTI_THREAD ) {
	       System.out.println("\nend executeFormatTask format\n");
	   }
       }
       endFormat();
   }

   protected final void executeBuildTask(ThreadConditionInterface aCondition) {
       PathInterface theUpdatePath = null;
       boolean theBuildCompleteFlag = false;
       synchronized ( this ) {
	   Vector theUpdatePathQueue = getBuildPathQueue();
	   if ( ! theUpdatePathQueue.isEmpty() ) {
	       theUpdatePath = (PathInterface)(theUpdatePathQueue.elementAt(0));
	       theUpdatePathQueue.removeElementAt(0);
	   }
	   else {
	       if ( _buildFrozen ) {
		   theBuildCompleteFlag = true;
	       }
	   }
       }
       if ( theUpdatePath != null ) {
	   theUpdatePath.accept(_pathBuildingVisitor,
				getRootGlyph(),
				getSelectionPath());
	   StackInterface theEmptyLeaves = _pathBuildingVisitor.getEmptyLeaves();
	   while ( ! theEmptyLeaves.isEmpty() ) {
	       ((SelectionPathInterface)theEmptyLeaves.pop()).cutEmptyLeaf();
	   }
	   getContext().buildAdvance();
       }
       else {
	   if ( theBuildCompleteFlag ) {
	       getContext().buildComplete();
	       endBuild();
	   }
	   else {
	       aCondition.await(); /// TEMPO TEMPO
	   }
       }
   }

   public final void initDone() {
       if ( Globals.DEBUG_MULTI_THREAD ) {
	   System.out.println("-> initDone");
       }
       synchronized ( this ) {
	   _threadState.init();
       }
       if ( Globals.DEBUG_MULTI_THREAD ) {
	   System.out.println("<- initDone");
       }
       awake();
   }

   public final void fatalError() {
       if ( Globals.DEBUG_MULTI_THREAD ) {
	   System.out.println("-> fatalError");
       }
       synchronized ( this ) {
	   _threadState.fatalError();
       }
       if ( Globals.DEBUG_MULTI_THREAD ) {
	   System.out.println("<- fatalError");
       }
       awake();
       throw new ImplementationError("FatalError");
   }
    
   public final void disposeWhenInactive() {
       if ( Globals.DEBUG_MULTI_THREAD ) {
	   System.out.println("-> disposeWhenInactive");
       }
       try {
	   synchronized ( _inactive ) {
	       while ( true ) {
		   synchronized ( this ) {
		       if ( _threadState.isInactive() ) {
			   _threadState.quit();
			   break;
		       }
		   }
		   _inactive.await();
	       }
	   }
	   awakeBuilding();
	   awakeFormatting();
	   awake();
	   dispose();
       }
       catch ( StoppedFacadeException anException ) {
	   return;
       }
       if ( Globals.DEBUG_MULTI_THREAD ) {
	   System.out.println("<- disposeWhenInactive");
       }
   }

   protected final void finalize() throws Throwable {
       if ( ! _disposedFlag ) {
	   disposeWhenInactive();
       }
       super.finalize();
   }

   public final boolean isDisposed() {
       return _disposedFlag;
   }

   protected void dispose() {
       _disposedFlag = true;
       _version = null;
       _factory.dispose();
       _factory = null;
       _rootGlyph.dispose();
       _rootGlyph = null;
       _selections = null;
       _rootSelectionPath = null;
       _buildPathQueue = null;
       _rectangleCacheForRefresh = null;
       _booleanCacheForRefresh = null;
       _rectangleCacheForRefreshVisible = null;
       _rectangleCacheForRefreshWork = null;
       _emptyRectangle = null;
       _alternative.dispose();
       _alternative = null;
       _pathModificationVisitor.dispose();
       _pathModificationVisitor = null;
       _pathBuildingVisitor.dispose();
       _pathBuildingVisitor = null;
       _distance = null;
       _formatable = null;
       _formatSuspended = null;
       _drawable = null;
       _building = null;
       _inactive = null;
       _formatThread = null;
       _buildThread = null;
       _pathSelectionVisitor.dispose();
       _pathSelectionVisitor = null;
       _errorHandler = null;
       _threadState.dispose();
       _threadState = null;
   }

    public final void writeAsXML(Writer anOutput) throws IOException {
	anOutput.write("<?XML version=\"1.0\"?>");
	writeState(anOutput);
    }

    public final void writeState(Writer anOutput) throws IOException {
	anOutput.write("<Facade>");
	anOutput.write("<Resources>");
	_factory.writeState(anOutput);
	anOutput.write("</Resources>");
	String theKey;
	boolean theFirstFlag = true;
	Enumeration theIterator = _selections.keys();
	while ( theIterator.hasMoreElements() ) {
	    if ( theFirstFlag ) {
		theFirstFlag = false;
		anOutput.write("<Selections>");
	    }
	    theKey = (String)theIterator.nextElement();
	    ((SelectionInterface)_selections.get(theKey)).writeState(anOutput);
	}
	if ( ! theFirstFlag ) {
	    anOutput.write("</Selections>");
	}
	_rootGlyph.writeState(anOutput);
	anOutput.write("</Facade>");
    }

   private boolean _disposedFlag;
   private int _formattingWidth;
   private AvailableVersion _version;
   private ResourceFactoryInterface _factory;
   private RootGlyph _rootGlyph;
   private Hashtable _selections;
   private SelectionPathInterface _rootSelectionPath;
   private Vector _buildPathQueue;
   private boolean _buildFrozen;
   private Rectangle _rectangleCacheForRefresh;
   private BooleanWrapper _booleanCacheForRefresh;
   private Rectangle _rectangleCacheForRefreshVisible;
   private Rectangle _rectangleCacheForRefreshWork;
   private Rectangle _emptyRectangle;
   private AlternativeInterface _alternative;
   private PathModificationVisitor _pathModificationVisitor;
   private PathBuildingVisitor _pathBuildingVisitor;
   private ImmediateDistance _distance;
   private NotifyingThreadStateInterface _threadState;
   private ThreadConditionInterface _formatable;
   private ThreadConditionInterface _formatSuspended;
   private ThreadConditionInterface _drawable;
   private ThreadConditionInterface _building;
   private ThreadConditionInterface _inactive;
   private Thread _formatThread;
   private Thread _buildThread;
   private PathSelectionVisitor _pathSelectionVisitor;
   private ErrorHandlerInterface _errorHandler;

}

final class RootSelectionPath extends SelectionPath {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/Facade.java,v $ $revision$ $Date: 1998/11/02 15:54:28 $ Copyright 1998 INRIA.";
    
    protected RootSelectionPath() {
	super(0);
    }
    public final void removeAllSelections(PathVisitorInterface aVisitor) {
	removeSelections();
    }
}



