//*****************************************************************************
//   Copyright (c) 1998 INRIA.
//   http://www.inria.fr/croap/aioli/modules/figue/web/credits.html#COPYRIGHT
//*****************************************************************************

package figue.internal;

import figue.custom.ConfigurationInterface;

import figue.path.SelectionPathInterface;
import figue.resource.ResourceFactoryInterface;
import figue.resource.ASCIIResourceFactory;
import figue.resource.AWTResourceFactory;
import figue.resource.FontMetricInterface;
import figue.geometry.Transformer;
import figue.geometry.Rectangle;
import figue.format.DistanceInterface;
import figue.format.AlternativeInterface;
import figue.format.StrategySetOrigin;
import figue.format.StrategySetSize;
import figue.format.StrategyImproveSizeEstimation;
import figue.format.StrategyReadyToFormat;
import figue.format.StopFormattingSonsException;
import figue.error.ErrorHandlerInterface;
import figue.error.ImplementationError;
import figue.box.GlyphInterface;

import java.net.URL;

public abstract class AvailableVersion {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/AvailableVersion.java,v $ $revision$ $Date: 1998/11/02 15:55:11 $ Copyright 1998 INRIA.";

    AvailableVersion(String aName,
		     ContextInterface aContext) {
	_name = aName;
	_context = aContext;
	_strategySetOriginIncremental = StrategySetOrigin.INCREMENTAL.clone(aContext);
	_strategySetOriginNoIncremental = StrategySetOrigin.NO_INCREMENTAL.clone(aContext);
	_strategySetSizeIncremental = StrategySetSize.INCREMENTAL.clone(aContext);
	_strategySetSizeNoIncremental = StrategySetSize.NO_INCREMENTAL.clone(aContext);
    }

    AvailableVersion(String aName) {
	this(aName,null);
    }

    AvailableVersion(AvailableVersion aVersion,
		     ContextInterface aContext) {
	_name = aVersion._name;
	_context = aContext;
	_strategySetOriginIncremental = aVersion._strategySetOriginIncremental.clone(aContext);
	_strategySetOriginNoIncremental = aVersion._strategySetOriginNoIncremental.clone(aContext);
	_strategySetSizeIncremental = aVersion._strategySetSizeIncremental.clone(aContext);
	_strategySetSizeNoIncremental = aVersion._strategySetSizeNoIncremental.clone(aContext);
    }

    protected final ContextInterface getContext() {
	return _context;
    }
    public final void setContext(ContextInterface aContext) {
	_context = aContext;
	_strategySetOriginIncremental.setContext(aContext);
	_strategySetOriginNoIncremental.setContext(aContext);
	_strategySetSizeIncremental.setContext(aContext);
	_strategySetSizeNoIncremental.setContext(aContext);
    }
    protected final StrategySetOrigin getStrategySetOriginIncremental() {
	return _strategySetOriginIncremental;
    }
    protected final StrategySetOrigin getStrategySetOriginNoIncremental() {
	return _strategySetOriginNoIncremental;
    }
    protected final StrategySetSize getStrategySetSizeIncremental() {
	return _strategySetSizeIncremental;
    }
    protected final StrategySetSize getStrategySetSizeNoIncremental() {
	return _strategySetSizeNoIncremental;
    }

    public abstract ResourceFactoryInterface newResourceFactory(ErrorHandlerInterface anErrorHandler,URL aPortableFontUrl);

    public abstract int getDefaultFormattingWidth();

    public abstract void format(GlyphInterface aGlyph,
				FontMetricInterface aFontMetric,
				AlternativeInterface anAlternative,
				DistanceInterface aDistance);
    
    public abstract void reFormat(GlyphInterface aGlyph,
				  FontMetricInterface aFontMetric,
				  AlternativeInterface anAlternative,
				  DistanceInterface aDistance);

    public abstract void draw(GlyphInterface aGlyph,
			      Transformer aTransformer,
			      SelectionPathInterface aSelectionPath,
			      Rectangle aRefreshZone,
			      Rectangle aBoundingBox);
    
    public abstract AvailableVersion clone(ContextInterface aContext);
  
    public final boolean equals(Object anObject) {
	if ( this == anObject ) {
	    return true;
	}
	if ( getClass() != anObject.getClass() ) {
	    return false;
	}
	AvailableVersion theVersion = (AvailableVersion)anObject;
	return _name.equals(theVersion._name);
    }
  
    public final String toString() {
	StringBuffer theResult = new StringBuffer(_name);
	theResult.append("Version");
	return theResult.toString();
    }

    private String _name;
    private ContextInterface _context;
    private StrategySetOrigin _strategySetOriginIncremental;
    private StrategySetOrigin _strategySetOriginNoIncremental;
    private StrategySetSize _strategySetSizeIncremental;
    private StrategySetSize _strategySetSizeNoIncremental;

    public final static AvailableVersion GRAPHICAL = new GraphicalVersion();
    public final static AvailableVersion PRINTING = new PrintingVersion();
    public final static AvailableVersion TEXTUAL = new TextualVersion();
  
}

final class GraphicalVersion extends AvailableVersion {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/AvailableVersion.java,v $ $revision$ $Date: 1998/11/02 15:55:11 $ Copyright 1998 INRIA.";
  
    public GraphicalVersion() {
	super("Graphical");
    }

    private GraphicalVersion(GraphicalVersion aVersion,
			     ContextInterface aContext) {
	super(aVersion,aContext);
    }

    public final AvailableVersion clone(ContextInterface aContext) {
	return new GraphicalVersion(this,aContext);
    }

    public final ResourceFactoryInterface newResourceFactory(ErrorHandlerInterface anErrorHandler,URL aPortableFontUrl) {
	return new AWTResourceFactory(anErrorHandler,aPortableFontUrl);
    }

    public final int getDefaultFormattingWidth() {
	return Globals.DEFAULT_FORMATTING_WIDTH_AWT;
    }

    public final void format(GlyphInterface aGlyph,
			     FontMetricInterface aFontMetric,
			     AlternativeInterface anAlternative,
			     DistanceInterface aDistance) {
	try {
	    aGlyph.format2D(getContext(),
			    aFontMetric,
			    anAlternative,
			    aDistance,
			    StrategyReadyToFormat.UNSAFE,
			    StrategyImproveSizeEstimation.ACTIVE,
			    getStrategySetOriginNoIncremental(),
			    getStrategySetSizeNoIncremental(),
			    false);
	    getContext().lastMajorFormatAdvance(); /// +++
	}
	catch ( StopFormattingSonsException anException ) {
	    throw new ImplementationError("Unexpected");
	}
    }

    public final void reFormat(GlyphInterface aGlyph,
			       FontMetricInterface aFontMetric,
			       AlternativeInterface anAlternative,
			       DistanceInterface aDistance) {
	try {
	    aGlyph.format2D(getContext(),
			    aFontMetric,
			    anAlternative,
			    aDistance,
			    StrategyReadyToFormat.SAFE,
			    StrategyImproveSizeEstimation.INACTIVE,
			    getStrategySetOriginIncremental(),
			    getStrategySetSizeIncremental(),
			    false);
	    getContext().lastMajorFormatInteractiveAdvance(); /// +++
	}
	catch ( StopFormattingSonsException anException ) {
	    throw new ImplementationError("Unexpected");
	}
    }

    public final void draw(GlyphInterface aGlyph,
			   Transformer aTransformer,
			   SelectionPathInterface aSelectionPath,
			   Rectangle aRefreshZone,
			   Rectangle aBoundingBox) {
	aGlyph.draw2D(getContext(),aTransformer,aSelectionPath,aRefreshZone,aBoundingBox);
    }
  
}

final class TextualVersion extends AvailableVersion {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/AvailableVersion.java,v $ $revision$ $Date: 1998/11/02 15:55:11 $ Copyright 1998 INRIA.";
  
    TextualVersion() {
	super("Textual");
    }

    private TextualVersion(TextualVersion aVersion,
			   ContextInterface aContext) {
	super(aVersion,aContext);
    }

    public final AvailableVersion clone(ContextInterface aContext) {
	return new TextualVersion(this,aContext);
    }
  
    public final ResourceFactoryInterface newResourceFactory(ErrorHandlerInterface anErrorHandler,URL aPortableFontUrl) {
	return new ASCIIResourceFactory(anErrorHandler,aPortableFontUrl);
    }

    public final int getDefaultFormattingWidth() {
	return Globals.DEFAULT_FORMATTING_WIDTH_ASCII;
    }

    public final void format(GlyphInterface aGlyph,
			     FontMetricInterface aFontMetric,
			     AlternativeInterface anAlternative,
			     DistanceInterface aDistance) {
	try {
	    aGlyph.format2D(getContext(),
			    aFontMetric,
			    anAlternative,
			    aDistance,
			    StrategyReadyToFormat.UNSAFE,
			    StrategyImproveSizeEstimation.INACTIVE,
			    getStrategySetOriginNoIncremental(),
			    getStrategySetSizeNoIncremental(),
			    false);
	}
	catch ( StopFormattingSonsException anException ) {
	    throw new ImplementationError("Unexpected");
	}
    }

    public final void reFormat(GlyphInterface aGlyph,
			       FontMetricInterface aFontMetric,
			       AlternativeInterface anAlternative,
			       DistanceInterface aDistance) {
	format(aGlyph,aFontMetric,anAlternative,aDistance);
    }

    public final void draw(GlyphInterface aGlyph,
			   Transformer aTransformer,
			   SelectionPathInterface aSelectionPath,
			   Rectangle aRefreshZone,
			   Rectangle aBoundingBox) {
	aGlyph.draw2D(getContext(),aTransformer,aSelectionPath,aRefreshZone,aBoundingBox);
    }

}

final class PrintingVersion extends AvailableVersion {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/AvailableVersion.java,v $ $revision$ $Date: 1998/11/02 15:55:11 $ Copyright 1998 INRIA.";

    public PrintingVersion() {
	super("Printing");
    }

    private PrintingVersion(PrintingVersion aVersion,
			    ContextInterface aContext) {
	super(aVersion,aContext);
    }

    public final AvailableVersion clone(ContextInterface aContext) {
	return new PrintingVersion(this,aContext);
    }

    public final ResourceFactoryInterface newResourceFactory(ErrorHandlerInterface anErrorHandler,URL aPortableFontUrl) {
	return new AWTResourceFactory(anErrorHandler,aPortableFontUrl);
    }
  
    public final int getDefaultFormattingWidth() {
	return Globals.DEFAULT_FORMATTING_WIDTH_PS;
    }
  
    public final void format(GlyphInterface aGlyph,
			     FontMetricInterface aFontMetric,
			     AlternativeInterface anAlternative,
			     DistanceInterface aDistance) {
	try {
	    aGlyph.format2D(getContext(),
			    aFontMetric,
			    anAlternative,
			    aDistance,
			    StrategyReadyToFormat.UNSAFE,
			    StrategyImproveSizeEstimation.INACTIVE,
			    getStrategySetOriginNoIncremental(),
			    getStrategySetSizeNoIncremental(),
			    false);
	}
	catch ( StopFormattingSonsException anException ) {
	    throw new ImplementationError("Unexpected");
	}
    }

    public final void reFormat(GlyphInterface aGlyph,
			       FontMetricInterface aFontMetric,
			       AlternativeInterface anAlternative,
			       DistanceInterface aDistance) {
	format(aGlyph,aFontMetric,anAlternative,aDistance);
    }
  
    public final void draw(GlyphInterface aGlyph,
			   Transformer aTransformer,
			   SelectionPathInterface aSelectionPath,
			   Rectangle aRefreshZone,
			   Rectangle aBoundingBox) {
	aGlyph.draw2D(getContext(),aTransformer,aSelectionPath,aRefreshZone,aBoundingBox);
    }

}

