//*****************************************************************************
//   Copyright (c) 1998 INRIA.
//   http://www.inria.fr/croap/aioli/modules/figue/web/credits.html#COPYRIGHT
//*****************************************************************************

package figue.internal;

import figue.box.GlyphInterface;

import java.util.Enumeration;
import java.util.NoSuchElementException;

public final class VectorGlyph implements Cloneable {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/VectorGlyph.java,v $ $revision$ $Date: 1998/11/02 15:55:18 $ Copyright 1998 INRIA.";
  
    protected GlyphInterface elementData[];
    protected int elementCount;
    protected int capacityIncrement;
  
    public VectorGlyph(int initialCapacity, int capacityIncrement) {
	super();
	this.elementData = new GlyphInterface[initialCapacity];
	this.capacityIncrement = capacityIncrement;
    }

    public VectorGlyph(int initialCapacity) {
	this(initialCapacity, 0);
    }

    public VectorGlyph() {
	this(10);
    }

    public final synchronized void copyInto(GlyphInterface anArray[]) {
	int i = elementCount;
	while (i-- > 0) {
	    anArray[i] = elementData[i];
	}
    }

    public final synchronized void trimToSize() {
	int oldCapacity = elementData.length;
	if (elementCount < oldCapacity) {
	    GlyphInterface oldData[] = elementData;
	    elementData = new GlyphInterface[elementCount];
	    System.arraycopy(oldData, 0, elementData, 0, elementCount);
	}
    }

    public final synchronized void ensureCapacity(int minCapacity) {
	int oldCapacity = elementData.length;
	if (minCapacity > oldCapacity) {
	    GlyphInterface oldData[] = elementData;
	    int newCapacity = (capacityIncrement > 0) ?
		(oldCapacity + capacityIncrement) : (oldCapacity * 2);
	    if (newCapacity < minCapacity) {
		newCapacity = minCapacity;
	    }
	    elementData = new GlyphInterface[newCapacity];
	    System.arraycopy(oldData, 0, elementData, 0, elementCount);
	}
    }

    public final synchronized void setSize(int newSize) {
	if (newSize > elementCount) {
	    ensureCapacity(newSize);
	} else {
	    for (int i = newSize ; i < elementCount ; i++) {
		elementData[i] = null;
	    }
	}
	elementCount = newSize;
    }

    public final int capacity() {
	return elementData.length;
    }
  
    public final int size() {
	return elementCount;
    }
  
    public final boolean isEmpty() {
	return elementCount == 0;
    }
  
    public final synchronized Enumeration elements() {
	return new VectorGlyphEnumerator(this);
    }
  
    public final boolean contains(GlyphInterface elem) {
	return indexOf(elem, 0) >= 0;
    }

    public final int indexOf(GlyphInterface elem) {
	return indexOf(elem, 0);
    }

    public final synchronized int indexOf(GlyphInterface elem, int index) {
	for (int i = index ; i < elementCount ; i++) {
	    if (elem.equals(elementData[i])) {
		return i;
	    }
	}
	return -1;
    }

    public final int lastIndexOf(GlyphInterface elem) {
	return lastIndexOf(elem, elementCount-1);
    }
  
    public final synchronized int lastIndexOf(GlyphInterface elem, int index) {
	for (int i = index ; i >= 0 ; i--) {
	    if (elem.equals(elementData[i])) {
		return i;
	    }
	}
	return -1;
    }

    public final synchronized GlyphInterface elementAt(int index) {
	return elementData[index];
    }
  
    public final synchronized GlyphInterface firstElement() {
	return elementData[0];
    }
  
    public final synchronized GlyphInterface lastElement() {
	return elementData[elementCount - 1];
    }
  
    public final synchronized void setElementAt(GlyphInterface obj, int index) {
	elementData[index] = obj;
    }

    public final synchronized void removeElementAt(int index) {
	int j = elementCount - index - 1;
	if (j > 0) {
	    System.arraycopy(elementData, index + 1, elementData, index, j);
	}
	elementCount--;
	elementData[elementCount] = null;
    }

    public final synchronized void insertElementAt(GlyphInterface obj, int index) {
	ensureCapacity(elementCount + 1);
	System.arraycopy(elementData, index, elementData, index + 1, elementCount - index);
	elementData[index] = obj;
	elementCount++;
    }

    public final synchronized void addElement(GlyphInterface obj) {
	ensureCapacity(elementCount + 1);
	elementData[elementCount++] = obj;
    }

    public final synchronized boolean removeElement(GlyphInterface obj) {
	int i = indexOf(obj);
	if (i >= 0) {
	    removeElementAt(i);
	    return true;
	}
	return false;
    }

    public final synchronized void removeAllElements() {
	for (int i = 0; i < elementCount; i++) {
	    elementData[i] = null;
	}
	elementCount = 0;
    }

    public synchronized Object clone() throws CloneNotSupportedException {
	VectorGlyph v = (VectorGlyph)super.clone();
	v.elementData = new GlyphInterface[elementCount];
	for ( int theRank = 0; theRank < elementCount; theRank++ ) {
	    v.elementData[theRank] = (GlyphInterface)elementData[theRank].clone();
	}
	return v;
    }

//     public synchronized Object clone() {
// 	try { 
// 	    VectorGlyph v = (VectorGlyph)super.clone();
// 	    v.elementData = new GlyphInterface[elementCount];
// 	    System.arraycopy(elementData, 0, v.elementData, 0, elementCount);
// 	    return v;
// 	} catch (CloneNotSupportedException e) { 
// 	    throw new InternalError();
// 	}
//     }
  
    public final synchronized String toString() {
	int max = size() - 1;
	StringBuffer buf = new StringBuffer();
	Enumeration e = elements();
	buf.append("[");
    
	for (int i = 0 ; i <= max ; i++) {
	    String s = e.nextElement().toString();
	    buf.append(s);
	    if (i < max) {
		buf.append(", ");
	    }
	}
	buf.append("]");
	return buf.toString();
    }
}

final class VectorGlyphEnumerator implements Enumeration {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/VectorGlyph.java,v $ $revision$ $Date: 1998/11/02 15:55:18 $ Copyright 1998 INRIA.";

    VectorGlyph vector;
    int count;
  
    VectorGlyphEnumerator(VectorGlyph v) {
	vector = v;
	count = 0;
    }
  
    public boolean hasMoreElements() {
	return count < vector.elementCount;
    }
  
    public Object nextElement() {
	synchronized (vector) {
	    if (count < vector.elementCount) {
		return vector.elementData[count++];
	    }
	}
	throw new NoSuchElementException("VectorGlyphEnumerator");
    }
}
